/* Bartels LCS Spec Software V1.0 (17.08.2005) */

/* (C) 2005 Oliver Bartels F+E, Erding */

/* usbconnector.cpp : USB Connector class */

#include "stdafx.h"
#include <process.h>
#include "Lcs.h"

/* device driver */
#include "ftd2xx.h"
/* class declaration */
#include "usbconnector.h"
/* Interface declaration */
#include "USBCMD.H"
#include "LCSSETV.H"

/* handle access */
#define GETHANDLE (*((FT_HANDLE*) &m_handledata[0]))

/* USB constants */
#define USB_RDTIMEOUT	500		/* USB read timeout (ms) */
#define USB_WRTIMEOUT	1000		/* USB write timeout (ms) */
#define USB_IDTRIES	6			/* USB identification tries */
#define USB_PSTATTRIES	20		/* USB prog. status read tries */
#define USB_IDCODELEN	4		/* USB ident. code length */

/* global instance */
Cusbconnection * Cusbconnection::m_instance;
int Cusbconnection::m_abortflag = 0;

struct SCmdInfo {		/* command description */
	int cmd;
	int len;
	const char * name;
};

/****************************************************************************
*
* Constructor / Object Creation
*
****************************************************************************/
Cusbconnection::Cusbconnection()
/* Constructor */
{
	/* not open yet */
	m_state = 0;
	m_timeoutcount = 2;
	m_errorstate = NONE;
	memset(m_handledata, 0, sizeof(m_handledata));
	m_debuglen = -1;
	m_debugtxt[0] = '\0';
	m_bufstart = 0;
	m_bufend = 0;
	m_lastnoptime = 0;
	m_devicename[0] = '\0';
	/* reset reader thread */
	m_terminatereader = NULL;
	m_readerstate = 0;
	/* create thread sync mutex */
	m_mutex = CreateMutex(NULL, FALSE, NULL);
}

Cusbconnection::~Cusbconnection()
/* Destructor */
{
	/* terminate reader */
	if (m_terminatereader != NULL) {
		/* set terminate flag */
		*m_terminatereader = 1;
		m_terminatereader = NULL;
		/* wait a short time */
		Sleep(1);
		}
	/* shutdown mutex */
	CloseHandle(m_mutex);
	/* ensure closed state */
	doclose();
}

Cusbconnection * Cusbconnection::getconnection(const char * name)
/* get a connection by id (static) */
/* Returns: an with active connection or NULL */
/*
// const char * name	: serial number of the device
*/
{
	Cusbconnection * conn;			/* new connection */
	/* check: device is present */
	if (m_instance != NULL) {
		return m_instance;
		}
	/* create new connection */
	conn = new Cusbconnection();
	if (conn == NULL) {
		return NULL;
		}
	/* store connection */
	m_instance = conn;
	/* open connection */
	if (conn -> doopen(name) != 0) {
		return conn;
		}
	/* stop value reporter */
	conn -> writeenvironment(SV_USBMEAS, 0.0f);
	/* return the connection object */
	return conn;
}

void Cusbconnection::seterrorstate(int state, int code)
/* store error state */
/*
// int state		: state where error occured
// int code			: last error code
*/
{
	/* don't override first error */
	if (m_errorstate == NONE) {
		m_errorstate = state;
		m_errorcode = code;
		}
	else if (state == NONE) {
		m_errorstate = state;
		m_errorcode = 0;
		}
}

void Cusbconnection::abort()
/* set abort flag */
{
	m_abortflag = 1;
}

/****************************************************************************
*
* Connection management
*
****************************************************************************/
int Cusbconnection::doopen(const char * name)
/* open connection */
/* Returns: -1 on error, 0 on success */
/*
// const char * name	: serial number of the device
*/
{
	FT_STATUS status;				/* driver call state */
	DWORD devcnt;					/* number of devices */
	FT_DEVICE ftDevice;				/* device type (dummy) */
	DWORD deviceID;					/* device id (dummy) */
	char Description[64];			/* device description (dummy) */
	/* close first */
	doclose();
	/* check: no specific device requested */
	if ((name == NULL)
	|| (strlen(name) == 0)) {
		/* List all FTDI devices */
		status = FT_ListDevices(&devcnt, NULL, FT_LIST_NUMBER_ONLY);
		if (status != FT_OK) {
			seterrorstate(LISTDEVICES, status);
			return -1;
			}
		if (devcnt == 0) {
			seterrorstate(LISTDEVICES, status);
			return -1;
			}
		/* need device specification */
		else if (devcnt > 1) {
			seterrorstate(NOTUNIQUE, 0);
			return -1;
			}
		/* open the one and only device */
		memset(&m_handledata, 0, sizeof(m_handledata));
		status = FT_Open(0, &GETHANDLE);
		if (status != FT_OK) {
			seterrorstate(OPEN, status);
			return -1;
			}
		}
	/* specific device wanted */
	else {
		/* open the specified device */
		memset(&m_handledata, 0, sizeof(m_handledata));
		status = FT_OpenEx((void*) name, 
			FT_OPEN_BY_SERIAL_NUMBER, &GETHANDLE);
		if (status != FT_OK) {
			if (status ==  FT_DEVICE_NOT_FOUND) {
				seterrorstate(NOTFOUND, status);
				}
			else {
				seterrorstate(OPEN, status);
				}
			return -1;
			}
		}
	/* read the device name */
	FT_GetDeviceInfo(GETHANDLE, &ftDevice, &deviceID, 
		m_devicename, Description, NULL);
	/* Perform the USB timeout setup */
	status = FT_SetTimeouts(GETHANDLE, USB_RDTIMEOUT, USB_WRTIMEOUT);
	if (status != FT_OK) {
		seterrorstate(SETUPDEVICE, status);
		doclose();
		return -1;
		}
	/* Perform the USB reset retry setup */
	status = FT_SetResetPipeRetryCount(GETHANDLE, 5);
	if (status != FT_OK) {
		seterrorstate(SETUPDEVICE, status);
		doclose();
		return -1;
		}
	/* Perform the USB flow control */
	status = FT_SetFlowControl(GETHANDLE, FT_FLOW_NONE, 0, 0);
	if (status != FT_OK) {
		seterrorstate(SETUPDEVICE, status);
		doclose();
		return -1;
		}
	status = FT_SetDataCharacteristics(GETHANDLE, 
		FT_BITS_8,FT_STOP_BITS_2,FT_PARITY_NONE);
	FT_SetBaudRate(GETHANDLE, FT_BAUD_921600);
	if (status != FT_OK) {
		seterrorstate(SETUPDEVICE, status);
		doclose();
		return -1;
		}
	FT_SetLatencyTimer(GETHANDLE, 1);
	FT_SetChars(GETHANDLE, 0, 0, 0, 0);
	FT_SetUSBParameters(GETHANDLE, 0x1000, 0x000);
	/* we're connected now */
	m_state = 1;
	/* check device id */
	if (verifyid() != 0) {
		seterrorstate(CHECKDEVICE, status);
		doclose();
		m_state = 0;
		return -1;
		}
	/* reset data input buffer */
	m_bufstart = 0;
	m_bufend = 0;
	/* done fine */
	return 0;
}

int Cusbconnection::doclose()
/* close connection */
/* Returns: -1 on error, 0 on success */
{
	FT_STATUS status;				/* driver call state */
	/* check: we're open, close */
	status = FT_OK;
	if (m_state != 0) {
		/* write stop data cmd */
		writecmdwords(UCMD_SETSTPVAR, 0x20, 0, 0);
		/* write logout cmd */
		writebytes(UCMD_LOGOUT);
		/* close the handle */
		status = FT_Close(GETHANDLE);
		}
	memset(m_handledata, 0, sizeof(m_handledata));
	/* done fine */
	if (status == FT_OK) {
		return 0;
		}
	seterrorstate(CLOSE, status);
	return -1;
}

void Cusbconnection::closeall()
/* close all connections */
{
	if (m_instance != NULL) {
		m_instance -> doclose();
		delete m_instance;
		m_instance = NULL;
		}
}

int Cusbconnection::verifyid(int force)
/* verfiy gadget id */
/* Returns: -1 on error, 0 on success */
/*
// int force	: flag: verify id no matter if already verifies
*/
{
	int i;						/* loop control */
	int oldstate = m_state;		/* current device state */
	int oldtimeout = m_timeoutcount;	/* current timeout */
	int success = 0;			/* success flag */
	static unsigned char idrequest[] = {	/* gadget id request code */
		UCMD_IDREQ, UIDREQ_CODE1, UIDREQ_CODE2, UIDREQ_CODE3, UIDREQ_CODE4 
		};
	unsigned char idreply[USB_IDCODELEN];	/* reply code from id */
	/* check: not open yet */
	if (m_state <= 0) {
		seterrorstate(CHECKDEVICE, 0);
		return -1;
		}
	/* check: already verified */
	if ((m_state >= 2)
		&& (force == 0)) {
		return 0;
		}
	/* set state to ok to enable read / write */
	m_state = 2;
	/* set low timeout for id verification */
	m_timeoutcount = 0;
	FT_SetTimeouts(GETHANDLE, USB_RDTIMEOUT / 5, USB_WRTIMEOUT);
	/* loop id check */
	for (i = 0; i < USB_IDTRIES; i++) {
		/* reset error state */
		seterrorstate(NONE, 0);
		/* write ID Request byte */
		if (write(idrequest, sizeof(idrequest)) == sizeof(idrequest)) {
			/* dummy byte due to communication problems */
			writebytes(UCMD_NOP);
			/* duplicate first request */
			if (i < 0) {
				if (write(idrequest, sizeof(idrequest)) == sizeof(idrequest)) {
					continue;
					}
				}
			/* read command */
			if (readcmd(idreply, UST_IDACK, USB_IDCODELEN) == USB_IDCODELEN) {
				if ((idreply[0] == UIDACK_CODE1)
					&& (idreply[1] == UIDACK_CODE2)
					&& (idreply[2] == UIDACK_CODE3)
					&& (idreply[3] == UIDACK_CODE4)) {
					/* good reply, id sync complete */
					success = 1;
					break;
					}
				}
			}
		else {
			Sleep(10);
			}
		}
	/* restore timeout flags */
	m_timeoutcount = oldtimeout;
	FT_SetTimeouts(GETHANDLE, USB_RDTIMEOUT, USB_WRTIMEOUT);
	/* report success */
	if (success != 0) {
		return 0;
		}
	/* not successfull, restore state and set error */
	m_state = oldstate;
	seterrorstate(CHECKDEVICE, 0);
	return -1;
}

void Cusbconnection::setdebug(int flag)
/* activate debugging */
/*
// int flag	: -1: clear current and reset, 
				0: dont debug
				1: debug
*/
{
	if (flag < 0) {
		m_debuglen = -1;
		m_debugtxt[0] = '\0';
		}
	else if (flag == 0) {
		m_debuglen = -1;
		}
	else {
		m_debuglen = strlen(m_debugtxt);
		}
}

const char * Cusbconnection::getdebug()
/* get debug text */
/* Returns: current debug text */
{
	return m_debugtxt;
}

const char * Cusbconnection::getdevicename()
/* get name of connected device */
/* Returns: the connected device or an empty string */
{
	if (m_state == 2) {
		return m_devicename;
		}
	return "";
}


/****************************************************************************
*
* Data input / output
*
****************************************************************************/
int Cusbconnection::read(void * output, int len, int block)
/* read data */
/* Returns: number of bytes written, -1 on error */
/*
// void * output	: target storage
// int len			: number of bytes to read
// int block		: alway wait until len is read
*/
{
	FT_STATUS status;				/* driver call state */
	DWORD readcount;				/* number of bytes read */
	int readlen = 0;				/* number of bytes written */
	int timeoutcount = 0;			/* number of timeouts */
	/* check memory buffer */
	if ((output == NULL)
	 || (len < 0)
	 || (m_state != 2)) {
		return -1;
		}
	/* continue to read buffer */
	while (readlen < len) {
		/* check abort */
		if (m_abortflag) {
			m_abortflag = 0;
			seterrorstate(READ, 0);
			return -1;
			}
		/* check status */
		if (m_state == 0) {
			seterrorstate(READ, 0);
			return -1;
			}
		/* try to read data */
		readcount = 0;
		status = FT_Read(GETHANDLE, &((unsigned char*) output)[readlen], 
			len, &readcount);
		if (status != FT_OK) {
			seterrorstate(READ, status);
			return -1;
			}
		/* check length */
		if (readcount > 0) {
			readlen += readcount;
			if (readlen > len) {
				return -1;
				}
			timeoutcount = 0;
			}
		else if ((++timeoutcount > m_timeoutcount)
			&& (block < 2)) {
			seterrorstate(READTIMEOUT, 0);
			return -1;
			}
		/* check: don't wait */
		else if (block == 0) {
			break;
			}
		}
	/* return number of bytes read */
	return readlen;
}

int Cusbconnection::readcmd(void * output, int expectedcmd, int len)
/* read specific command */
/* Returns: length of data in output, -1 on error */
/*
// void * output	: target buffer for the rest of the command
// int expectedcmd	: expected command
// int len			: length of buffer, -1 for implicit len
*/
{
	unsigned char cmd;				/* command read */
	unsigned char readcmdlen;		/* length byte of read command */
	int cmdlen;						/* length of cmd */
	unsigned char localbuf[256];	/* read buffer for other cmd */
	unsigned char * readbuf;		/* pointer to read buffer */
	int readbuflen;					/* length of readbuf */
	int syncsecuretime;				/* first time of bad sync */
	int syncsecurecounter = 0;		/* sync secure */
	/* continous reading */
	while (1) {
		/* read command */
		if (read(&cmd, 1) < 0) {
			return -1;
			}
		cmdlen = getcmdlen(cmd, -1);
		/* check: it's data, read to internal buffer */
		if (cmd == UST_AMPFREQDAT) {
			/* check: buffer free */
			int nextbuf = (m_bufend + 1) % MAXDATABUFFER;
			if (m_bufstart != nextbuf) {
				/* write to buffer */
				readbuf = (unsigned char *) &m_valbuf[m_bufend];
				readbuflen = sizeof(Smeasurementvalue);
				m_bufend = nextbuf;
				}
			/* drop data */
			else {
				readbuf = localbuf;
				readbuflen = sizeof(localbuf);
				}
			/* don't read endless data, the normal answer has to arrive in
			at least one second */
			if (cmd != expectedcmd) {
				if (syncsecurecounter == 0) {
					syncsecuretime = time(NULL);
					}
				else if (syncsecurecounter % 10 == 0) {
					if ((time(NULL) - syncsecuretime >= 2)
					|| ((time(NULL) - syncsecuretime >= 1)
					&& (syncsecurecounter > 500))) {
						if (expectedcmd != UST_IDACK) {
							seterrorstate(CMDSYNCWAIT, 0);
							return -1;
							}
						else if (time(NULL) - syncsecurecounter >= 2) {
							seterrorstate(CMDSYNCWAIT, 0);
							return -1;
							}
						}
					}
				syncsecurecounter++;
				}
			}
		/* expected command, set read length and buffer */
		else if (expectedcmd == cmd) {
			readbuf = (unsigned char *) output;
			readbuflen = len;
			}
		/* read to local buffer and drop data */
		else {
			readbuf = localbuf;
			readbuflen = sizeof(localbuf);
			}
		/* read length in case of data */
		if (cmdlen < 0) {
			if (cmd == UST_FTREADDAT) {
				if (read(&readcmdlen, 1) != 1) {
					seterrorstate(READCMDLEN, 0);
					return -1;
					}
				cmdlen = readcmdlen * 2;
				}
			/* use info from reader */
			else if ((cmd == expectedcmd)
				&& (len >= 0)) {
				cmdlen = len;
				}
			/* expect id sync, ignore other codes */
			else if (expectedcmd == UST_IDACK) {
				cmdlen = 0;
				}
			else {
				seterrorstate(READCMDCODE, 0);
				return -1;
				}
			}
		/* read the rest of the data */
		if (cmdlen > 0) {
			/* check buffer overflow */
			if ((cmdlen > readbuflen)
			&& (readbuflen >= 0)) {
				seterrorstate(READCMDLEN, 0);
				return -1;
				}
			/* read data */
			readbuflen = read(readbuf, cmdlen);
			if (readbuflen < 0) {
				seterrorstate(READ, readbuflen);
				return -1;
				}
			}
		/* check: expected command reached */
		if (cmd == expectedcmd) {
			return cmdlen;
			}
		}
	/* done fine */
	return 0;
}


int Cusbconnection::write(const void * data, int len)
/* write data */
/* Returns: -1 on error, number of bytes written else */
/*
// const void * data	: data to write
// int len				: number of bytes to write
*/
{
	FT_STATUS status;				/* driver call state */
	DWORD writecount;				/* number of bytes written by call */
	int writtenlen = 0;				/* number of bytes written */
	/* debug data */
	if (m_debuglen >= 0) {
		for (int i = 0; i < len; i++) {
			if (m_debuglen + 3 < sizeof(m_debugtxt)) {
				m_debuglen += sprintf(&m_debugtxt[m_debuglen], "%02X", 
					((const unsigned char*) data)[i]);
				m_debugtxt[m_debuglen] = '\0';
				}
			}
		}
	/* check connection state */
	if (m_state != 2) {
		return -1;
		}
	/* ensure, reader is stopped */
	if (pausereader() != 0) {
		return -1;
		}
	/* multiple call for partial transfers */
	while (writtenlen < len) {
		/* check abort */
		if (m_abortflag) {
			m_abortflag = 0;
			seterrorstate(WRITE, 0);
			return -1;
			}
		/* check status */
		if (m_state == 0) {
			seterrorstate(WRITE, 0);
			return -1;
			}
		/* try to write data */
		writecount = 0;
		status = FT_Write(GETHANDLE, 
			(void*) &((const unsigned char*) data)[writtenlen], 
			len - writtenlen, &writecount);
		if (status != FT_OK) {
			seterrorstate(WRITE, status);
			return -1;
			}
		/* check length */
		writtenlen += writecount;
		if (writtenlen > len) {
			return -1;
			}
		}
	/* done fine */
	return writtenlen;
}

int Cusbconnection::writebytes(int b1, int b2, int b3, int b4, int b5)
/* Write variabl number of bytes without buffer  */
/* Returns: -1 on error, 0 on success */
/*
// int b1	: byte 1
// int b2	: byte 2
// int b3	: byte 3
// int b4	: byte 4
// int b5	: byte 5
*/
{
	unsigned char buf[5];		/* buffer to write */
	int len = 0;				/* length of buffer */
	/* collect data */
	if (b1 >= 0) {
		buf[0] = b1;
		len++;
		if (b2 >= 0) {
			buf[1] = b2;
			len++;
			if (b3 >= 0) {
				buf[2] = b3;
				len++;
				if (b4 >= 0) {
					buf[3] = b4;
					len++;
					if (b5 >= 0) {
						buf[4] = b5;
						len++;
						}
					}
				}
			}
		}
	/* write data */
	return write(buf, len);
}

int Cusbconnection::writecmdwords(int cmd, int b1, int b2, int b3)
/* write cmd byte and b1-b3 as unsigned short */
/* Returns: -1 on error, 0 on success */
/*
// int cmd	: command byte
// int b1	: command parameter word 1
// int b2	: command parameter word 2
// int b3	: command parameter word 3
*/
{
	unsigned char buf[7];		/* buffer to write */
	int len;					/* length of buffer */
	if (cmd < 0) {
		return 0;
		}
	/* store command */
	buf[0] = cmd;
	len = 1;
	/* store data */
	if (b1 >= 0) {
		*((unsigned short *) &buf[1]) = (unsigned short) b1;
		len += 2;
		if (b2 >= 0) {
			*((unsigned short *) &buf[3]) = (unsigned short) b2;
			len += 2;
			if (b3 >= 0) {
				*((unsigned short *) &buf[5]) = (unsigned short) b3;
				len += 2;
				}
			}
		}
	/* write data */
	return write(buf, len);
}


int Cusbconnection::writeuint(const unsigned int data)
{
	return write(&data, sizeof(data));
}

int Cusbconnection::writeushort(const unsigned short data)
{
	return write(&data, sizeof(data));
}

int Cusbconnection::writeuchar(const unsigned char data)
{
	return write(&data, sizeof(data));
}

int Cusbconnection::readenvironment(int id, float & value)
/* read environment value */
/* Returns: -1 on error, 0 on success */
/*
// int id			: id of the variable
// float & value	: target value
*/
{
	unsigned char reply[5];		/* the reply */
	/* init value */
	value = 0.0f;
	if (writecmdwords(UCMD_GETSTPVAR, id) != 3) {
		return -1;
		}
	if (readcmd(&reply, UST_GETSTPRES, sizeof(reply)) < 0) {
		return -1;
		}
	/* check code */
	if (reply[0] != USST_DONE) {
		return -1;
		}
	/* convert the value */
	memcpy(&value, &reply[1], sizeof(value));
	/* done fine */
	return 0;
}

int Cusbconnection::writeenvironment(int id, float value)
/* write environtment value */
/* Returns: -1 on error, 0 on success */
/*
// int id		: 
// float value	: 
*/
{
	unsigned short * valptr;	/* pointer to float value */
	unsigned char status;		/* status reply */
	/* point to value */
	valptr = (unsigned short*) &value;
	/* write the value */
	if (writecmdwords(UCMD_SETSTPVAR, id, valptr[0], valptr[1]) != 7) {
		return -1;
		}
	/* read answer */
	status = USST_FAILED;
	readcmd(&status, UST_SETSTPRES);
	/* check status */
	if (status != USST_DONE) {
		return -1;
		}
	/* done fine */
	return 0;
}


int Cusbconnection::setmaxtimeout(int milliseconds)
/* set timeout */
/* Returns:  */
/*
// int milliseconds	: 
*/
{
	return m_timeoutcount = milliseconds / USB_RDTIMEOUT;
}

int Cusbconnection::getcmdlen(int cmd, int len)
/* get command reply rest length by command or len */
/* Returns: length of cmd (not including cmd) or len if unknown */
/*
// int cmd	: command
// int len	: default length
*/
{
	switch (cmd) {
	/* USB NOP command */
	case UST_NOP	:
		return 0;
	/* USB identification ack. */
	case UST_IDACK:
		return USB_IDCODELEN;
	/* USB file handling status */
	case UST_FILESTAT:
		return 1;
	/* USB file transfer max. size */
	case UST_FTMAXSIZE:
		return 1;
	/* USB file transfer read data */
	case UST_FTREADDAT:
		return len;
	/* USB flash mode report */
	case UST_FLMODEREP:
		return 1;
	/* USB get setup var. result */
	case UST_GETSTPRES:
		return 5;
	/* USB set setup var. result */
	case UST_SETSTPRES:
		return 1;
	/* USB amplitude frequency data */
	case UST_AMPFREQDAT:
		return 16;
	/* USB flash status */
	case UST_FLSTAT:
		return 1;
		}
	/* unknwon cmd, return default length */
	return len;
}

#define ES(name) else if (m_errorstate == name) { errortype =  #name; }
const char * Cusbconnection::geterrorstate()
/* get error description */
/* Returns: Error State Description or NULL == no error */
{
	const char * errortype;
	static char errortext[64];
	errortype = "UNKNWON";
	if (m_errorstate == NONE) {
		return NULL;
		}
	ES(OPEN)
	ES(LISTDEVICES)
	ES(NOTUNIQUE)
	ES(NOTFOUND)
	ES(CHECKDEVICE)
	ES(SETUPDEVICE)
	ES(CLOSE)
	ES(READ)
	ES(READCMDCODE)
	ES(READCMDLEN)
	ES(CMDSYNCWAIT)
	ES(WRITE)
	ES(READTIMEOUT)
	/* print the error */
	_snprintf(errortext, sizeof(errortext), "%s (%d)", errortype, m_errorcode);
	errortext[sizeof(errortext) - 1] = '\0';
	/* return the error */
	return errortext;
}

void __cdecl readerthreadfunc(void * instance)
/* run reader thread */
/*
// void * instance	: pointer to class that runs the reader
*/
{
	int terminateflag;
	((Cusbconnection*) instance) -> runreaderthread(&terminateflag);
}

int Cusbconnection::runreaderthread(int * terminate)
/* run reader thread */
/* Returns: -1 on error, 0 on success */
/*
// int * terminate	: pointer to terminate variable
*/
{
	unsigned long bytesready;	/* number of bytes ready to read */
	int status;					/* input status */
	HWND timerwnd;				/* copy of the timer window */
	int timerid;				/* copy of the timer id */
	/* init terminate */
	*terminate = 0;
	m_terminatereader = terminate;
	timerwnd = NULL;
	timerid = 0;
	/* continue while terminate flag isn't set */
	while (*terminate == 0) {
		/* sync */
		WaitForSingleObject(m_mutex, 1000);
		/* check: paused, don't try to read */
		if (m_readerstate == 2) {
			/* check: multiple times, sleep longer */
			bytesready++;
			if (bytesready > 5) {
				Sleep(m_pollinterval * 5);
				}
			}
		/* check connected state */
		else if (m_state != 2) {
			}
		/* ready to read */
		else {
			bytesready = 0;
			/* query queue status */
			while (1) {
				status = FT_GetQueueStatus(GETHANDLE, &bytesready);
				/* check: bad status, terminate reader */
				if (status != FT_OK) {
					seterrorstate(READ, status);
					status = FT_Close(GETHANDLE);
					*terminate = 1;
					m_terminatereader = NULL;
					m_readerstate = 0;
					PostMessage(m_readerwnd, WM_IDLEUPDATECMDUI, 0, 0);
					break;
					}
				/* check: no data, go to sleep */
				if (bytesready == 0) {
					break;
					}
				/* read data to buffer */
				readcmd(NULL, UST_AMPFREQDAT);
				}
			}
		/* report buffer data */
		if (m_bufstart != m_bufend) {
			/* check: no timer yet, start timer */
			if ((timerid == 0)
			|| (m_readertimerid == 0)) {
				/* start the timer */
				timerid = SetTimer(m_readerwnd, 
					m_readertimermsgid, m_readerdelay, NULL);
				timerwnd = m_readerwnd;
				m_readertimerid = timerid;
				}
			}
		/* sync done */
		ReleaseMutex(m_mutex);
		/* wait for next period */
		Sleep(m_pollinterval);
		}
	/* stop the timer using stack local vars */
	if (timerid != 0) {
		KillTimer(timerwnd, timerid);
		}
	/* done fine */
	return 0;
}

int Cusbconnection::startreader(HWND wnd, int msgid, int pollintervall)
/* start reader */
/* Returns: -1 on error, 0 on success, -2: already started */
/*
// HWND wnd				: window handle to send data ready message to
// int msgid			: id of the WM_TIMER message
// int pollintervall	: number of milliseconds to pause
*/
{
	/* check: reader is running */
	if (m_readerstate != 0) {
		return -2;
		}
	/* check: mutex exists */
	if (m_mutex == NULL) {
		m_mutex = CreateMutex(NULL, FALSE, NULL);
		}
	if (m_mutex == NULL) {
		return -1;
		}
	/* copy data */
	m_readerwnd = wnd;
	m_readertimermsgid = msgid;
	m_pollinterval = pollintervall;
	/* start the thread */
	m_readerstate = 1;
	_beginthread(readerthreadfunc, 4096, this);
	/* init standart value */
	m_readerdelay = 100;
	m_readertimerid = 0;
	m_readerflushlen = MAXDATABUFFER / 2;
	/* done fine */
	return 0;
}

int Cusbconnection::stopreader()
/* stop reader */
/* Returns: -1 on error, 0 on success */
{
	/* check: reader was started */
	if (m_readerstate != 0) {
		if (m_terminatereader == NULL) {
			return -1;
			}
		/* reset the local timer id */
		m_readertimerid = 0;
		/* tell him to stop */
		m_readerstate = 0;
		*m_terminatereader = 1;
		m_terminatereader = NULL;
		Sleep(1);
		}
	/* done fine */
	return 0;
}

int Cusbconnection::pausereader()
/* pause reader (until resume), thread pauses as soon as data is written */
/* Returns: -1 on error, 0 on success */
{
	/* check: reader is running */
	if (m_readerstate == 1) {
		/* pause the reader */
		m_readerstate = 2;
		/* wait for the mutex */
		WaitForSingleObject(m_mutex, 1000);
		ReleaseMutex(m_mutex);
		}
	/* done fine */
	return 0;
}

int Cusbconnection::resumereader()
/* resume stopped reader */
/* Returns: -1 on error, 0 on success */
{
	/* allow reader to continue */
	if (m_readerstate == 2) {
		m_readerstate = 1;
		}
	/* done fine */
	return 0;
}

int Cusbconnection::getreaderstate()
/* check reader state */
/* Returns: 0: stopped, 1: running, 2: paused */
{
	if (this == NULL) {
		return 0;
		}
	/* done fine */
	return m_readerstate;
}

int Cusbconnection::readdata(Smeasurementvalue * buf, int count)
/* get buffered data */
/* Returns: -1 on error, number of values read else */
/*
// Smeasurementvalue * buf	: target buffer
// int count				: number of entries in buf
*/
{
	int len;		/* number of entries to read */
	/* synchronize */
	WaitForSingleObject(m_mutex, 1000);
	/* calculate number of entries to send */
	len = (m_bufend - m_bufstart + MAXDATABUFFER) % MAXDATABUFFER;
	/* limit to count */
	if (len > count) {
		len = count;
		}
	/* check: no data to send, stop timer */
	else if ((len == 0)
	&& (m_readertimerid != 0)) {
		KillTimer(m_readerwnd, m_readertimerid);
		m_readertimerid = 0;
		}
	/* don't copy beyond buffer */
	if (len + m_bufstart > MAXDATABUFFER) {
		len = MAXDATABUFFER - m_bufstart;
		}
	/* copy data */
	memcpy(buf, &m_valbuf[m_bufstart], len * sizeof(Smeasurementvalue));
	m_bufstart = (m_bufstart + len + MAXDATABUFFER) % MAXDATABUFFER;
	/* sync done */
	ReleaseMutex(m_mutex);
	/* return number of entries remaining */
	return len;
}
